/**
 * Copyright(c) 2016-2018 FUJITSU LIMITED All rights reserved.
 */
/*************************************************************
 * include
 *************************************************************/
#include <linux/kernel.h>
#include <linux/slab.h>
#include "bmcd_common.h"

/*************************************************************
 * global
 *************************************************************/
static struct bmcd_ipmi_sram_dump1 sram_buf_dump1;
static struct bmcd_ipmi_sram_dump2 sram_buf_dump2;

/*************************************************************
 * prototype
 *************************************************************/
static bmcd_return_t bmcd_dump_start(struct bmcd_dump_info *);

/*************************************************************
 * function
 *************************************************************/
/**
 * @brief bmcd_dump_main
 * @param ダンプデータ、ダンプファイル名
 * @retval BMCドライバ共通戻り値
 * @note ダンプデータ転送シーケンスを実行する
 */
bmcd_return_t bmcd_dump_ready(struct bmcd_dump_info *info)
{
	struct bmcd_ipmi_param_dump1 param;
	struct bmcd_ipmi_sram_dump1 *sram = &sram_buf_dump1;
	bmcd_return_t rtn, rtn_save;

	param.sram_size = 4 + info->namesize;
	sram->file_size = info->datasize;
	memcpy(sram->file_name, info->filename, info->namesize);
	param.sram_check_sum = bmcd_common_checksum((int *)sram, param.sram_size);

	while (1) {
		/* Dump upload ready notice コマンドを送信 */
		rtn = bmcd_ipmi_send(BMCD_CMD_DUMP_UPLOAD_READY, &param, sram);
		if (rtn != E_NORMALEND) {
			BMCD_LOG(BMCD_WARN, "bmcd_ipmi_send abnormal end, rtn[%d].\n", rtn);
			break;
		}
		BMCD_LOG(BMCD_INFO, "Dump upload ready notice complete.\n");

		/* Dump upload start コマンドを送信 */
		rtn = bmcd_dump_start(info);
		if (rtn != E_NORMALEND) {
			/* Dump upload ready noticeコマンドからやり直す */
			BMCD_LOG(BMCD_WARN, "bmcd_dump_start abnormal end, rtn[%d].\n", rtn);
			break;
		}
		BMCD_LOG(BMCD_INFO, "Dump upload start complete.\n");
		break;
	}

	rtn_save = rtn;

	/* Dump upload end コマンドを送信 */
	rtn = bmcd_ipmi_send(BMCD_CMD_DUMP_UPLOAD_END, NULL, NULL);
	if (rtn != E_NORMALEND) {
		BMCD_LOG(BMCD_WARN, "bmcd_ipmi_send abnormal end, rtn[%d].\n", rtn);
	}
	else {
		BMCD_LOG(BMCD_INFO, "Dump upload end complete.\n");
	}

	/* BMCD_CMD_DUMP_UPLOAD_READYまたはDump upload start コマンドでエラーが発生した場合、
     * そちらのエラーを採用する */
	if (rtn_save != E_NORMALEND) {
		rtn = rtn_save;
	}

	return rtn;
}

static bmcd_return_t bmcd_dump_start(struct bmcd_dump_info *info)
{
	bmcd_return_t rtn;
	struct bmcd_ipmi_sram_dump2 *sram = &sram_buf_dump2;

	char *upload_data = info->info;		/* アップロードデータ */
	uint32_t data_size = info->datasize;	/* アップロードデータの合計サイズ */
	uint32_t upload_size;			/* アップロード時のデータサイズ */
	uint32_t upload_pos = 0;		/* 現在のアップロード位置 */

	BMCD_LOG(BMCD_INFO, "Upload start, data_size[%u]\n", data_size);

	while (1) {
		/* アップロードするサイズを取得する */
		if ((data_size - upload_pos) > TX_BUFF_MAX_SIZE) {
			upload_size = TX_BUFF_MAX_SIZE;
		}
		else {
			upload_size = data_size - upload_pos;
		}

		memcpy(sram->upload_data, &(upload_data[upload_pos]), upload_size);

		BMCD_LOG(BMCD_INFO, "Uploading, upload_pos[%u] upload_size[%u]\n",
			upload_pos, upload_size);

		/**
 		 * Dump upload start コマンドを送信して、ダンプデータのアップロード
 		 * を要求する
 		 */
		rtn = bmcd_ipmi_send(BMCD_CMD_DUMP_UPLOAD_START, NULL, sram);
		if (rtn != E_NORMALEND) {
			BMCD_LOG(BMCD_WARN, "bmcd_ipmi_send abnormal end, rtn[%d].\n", rtn);
			break;
		}
		BMCD_LOG(BMCD_INFO, "bmcd_ipmi_send complete.\n");

		/**
 		 * 現在のアップロード位置を更新する
 		 * 全てのデータをアップロードした場合は、bmcd_dump_end()を呼び出して
 		 * Dump upload end コマンドを送信する
 		 */
		upload_pos += upload_size;
		if (upload_pos >= data_size) {
			BMCD_LOG(BMCD_INFO, "Upload end, upload_pos[%u] data_size[%u]\n",
				upload_pos, data_size);
			break;
		}
	}

	return rtn;
}

