/**
 * Copyright(c) 2016-2020 FUJITSU LIMITED All rights reserved.
 */

/*************************************************************
 * include
 *************************************************************/
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/ipmi.h>
#include <linux/delay.h>
#include <linux/timekeeping.h>
#include <linux/sched.h>
#include <linux/wait.h>
#include "bmcd_common.h"
#ifdef IPMI_STUB
#include "bmcd_ipmi_stub.h"
#endif

/*************************************************************
 * global
 *************************************************************/
extern int errno;
extern int panic_mode;

/* IPMIユーザ */
static ipmi_user_t ipmi_user = NULL;

/* IPMIコマンドとの非同期やり取りを同期的に処理するためのデータ */
struct ipmi_sync_param ipmi_sync_param;

/* IPMIコマンド応答時のコールバック関数 */
static void bmcd_ipmi_cmd_resp(struct ipmi_recv_msg *, void *);
static struct ipmi_user_hndl response_callback = {
	.ipmi_recv_hndl = bmcd_ipmi_cmd_resp,
};

/* IPMIコマンドのリトライ用メッセージバッファ */
static struct kernel_ipmi_msg ipmi_msg_buf;

/* IPMIコマンドのアドレッシングテーブル */
static struct ipmi_addr ipmi_addr_buf;

/* IPMIコマンドのリトライ回数とリトライ開始時間 */
static int retry_count;
static int retry_start_time;

/* IPMIコマンド異常時のパニック有無(パニックする=1、パニックしない=0) */
static int do_panic = 0;

/* IPMI Dump Upload Start(38:41)用シリアル番号 */
static u32 serial_no = 0;

/* ioremap関数にてマッピングされたSBCRAMのアドレス */
#ifdef QEMU
static void *sbcram_iomap_addr = 0;
#else
static void __iomem *sbcram_iomap_addr = 0;
#endif

#ifdef QEMU
extern char *ipmi_stub_get_sbcram_addr(void);
static char *sbcram_addr = 0;
#endif

#ifdef IPMI_STUB
#define __ipmi_create_user(if_num, handler, handler_data, user) \
	ipmi_stub_create_user(if_num, handler, handler_data, user)
#define __ipmi_destroy_user(user) ipmi_stub_destroy_user(user)
#define __ipmi_free_recv_msg(msg) ipmi_stub_free_recv_msg(msg)
#define __ipmi_request_settime(user,addr,msgid,msg,user_msg_data,priority,retries,retry_time_ms) \
	ipmi_stub_request_settime(user,addr,msgid,msg,user_msg_data,priority,retries,retry_time_ms)
#define __ipmi_poll_interface(user) ipmi_stub_poll_interface(user)

#else
#define __ipmi_create_user(if_num, handler, handler_data, user) \
	ipmi_create_user(if_num, handler, handler_data, user)
#define __ipmi_destroy_user(user) ipmi_destroy_user(user)
#define __ipmi_free_recv_msg(msg) ipmi_free_recv_msg(msg)
#define __ipmi_request_settime(user,addr,msgid,msg,user_msg_data,priority,retries,retry_time_ms) \
	ipmi_request_settime(user,addr,msgid,msg,user_msg_data,priority,retries,retry_time_ms)
#define __ipmi_poll_interface(user) ipmi_poll_interface(user)
#endif

/*************************************************************
 * define
 *************************************************************/
#define BMCD_IPMI_ENOMEM_RETRY_MAX	10			/* ENOMEM時のリトライ回数 */
#define BMCD_IPMI_RETRY 		0xFFFF			/* コマンドリトライ */
#define BMCD_IPMI_NOTRESP_TIME		(HZ * 200)		/* IPMIコマンドタイムアウト時間 */
#define BMCD_IPMI_NOTRESP_COUNT     200 * 1000 /* IPMIコマンドポーリング回数 */
#define BMCD_IPMI_SBCRAM_PHYADDR	0x0000000011000000UL	/* SBCRAM実アドレス */
#define BMCD_IPMI_SBCRAM_SIZE		64 * 1024		/* SBCRAMサイズ */
#define BMCD_IPMI_DATA_SIZE		16			/* IPMIのメッセージデータ最大サイズ */

/*************************************************************
 * function
 *************************************************************/
/**
 * @brief bmcd_ipmi_panic_set
 * @param BMC異常終了時の振る舞い
 *			1 : Panic
 *			0 : 何もしない(上位層に異常終了を返す)
 * @note BMCが異常終了したときの振る舞いを設定する
 */
void bmcd_ipmi_panic_set(int param)
{
	if ((param == 0) || (param == 1)) {
		do_panic = param;
	}
	else {
		BMCD_LOG(BMCD_WARN, "Invalid parameter. param[%d]\n", param);
	}
}

/**
 * @brief bmcd_ipmi_panic_get
 * @retval BMC異常終了時の振る舞い
 *			1 : Panic
 *			0 : 何もしない(上位層に異常終了を返す)
 * @note BMCが異常終了したときの振る舞いを取得する
 */
int bmcd_ipmi_panic_get(void)
{
	return do_panic;
}

static void bmcd_ipmi_data_set(struct kernel_ipmi_msg *msg, void *data, unsigned int len)
{
	static char buf[BMCD_IPMI_DATA_SIZE];

	if ((data == NULL) || (len == 0)) {
		msg->data = NULL;
		msg->data_len = 0;
	}
	else {
		memset(buf, 0, BMCD_IPMI_DATA_SIZE);
		memcpy(buf, data, len);
		msg->data = buf;
		msg->data_len = len;
	}
}

static int bmcd_ipmi_sram_set(char *data, unsigned int data_size)
{
	int write_size;
	int align_out_size;
	int pos;
#ifdef QEMU
	char *ram_addr = sbcram_iomap_addr;
#else
	uint32_t *val32;
	void __iomem *ram_addr = sbcram_iomap_addr;
#endif

	BMCD_LOG(BMCD_INFO, "data[%p], data_size[%d].\n", data, data_size);

	if (ram_addr == NULL) {
		BMCD_LOG(BMCD_WARN, "SBCRAM address is null.\n");
		return BMCD_ABNORMAL_END;
	}

	if (BMCD_IPMI_SBCRAM_SIZE < data_size) {
		write_size = BMCD_IPMI_SBCRAM_SIZE;
	}
	else {
		write_size = data_size;
	}

	/* 4バイト境界で書き込みしたときの余りのサイズを取得 */
	align_out_size = write_size % 4;

	/* 4バイト境界で書き込む */
	for (pos = 0; pos < (write_size - align_out_size); pos += 4) {
#ifdef QEMU
		memcpy(ram_addr + pos, &(data[pos]), 4);
		BMCD_LOG(BMCD_DEBUG, "pos[%d] data[%u] RAM[%u]\n",
			pos, (uint32_t)data[pos], (uint32_t)(*(ram_addr + pos)));
#else
		val32 = (uint32_t *)&data[pos];
		iowrite32(*val32, ram_addr + pos);
		BMCD_LOG(BMCD_DEBUG, "pos[%d] data[%u] RAM[%u]\n",
			pos, (uint32_t)data[pos], ioread32(ram_addr + pos));
#endif
	}

	/* 4バイト境界の余りのサイズを1バイトずつ書き込む */
	for (; pos < write_size; pos ++) {
#ifdef QEMU
		memcpy(ram_addr + pos, &(data[pos]), 1);
		BMCD_LOG(BMCD_DEBUG, "pos[%d] data[%u] RAM[%u]\n",
			pos, (uint8_t)data[pos], (uint8_t)(*(ram_addr + pos)));
#else
		iowrite8((uint8_t)data[pos], ram_addr + pos);
		BMCD_LOG(BMCD_DEBUG, "pos[%d] data[%u] RAM[%u]\n",
			pos, (uint8_t)data[pos], ioread8(ram_addr + pos));
#endif
	}

#ifdef QEMU
	sbcram_addr = ipmi_stub_get_sbcram_addr();
	if (sbcram_addr) {
		memcpy(sbcram_addr, ram_addr, BMCD_IPMI_SBCRAM_SIZE);
	}
#endif

	BMCD_LOG(BMCD_INFO, "sbcram_iomap_addr[%p] write_size[%d], pos[%d].\n", 
		sbcram_iomap_addr, write_size, pos);
	return BMCD_NORMAL_END;
}

/**
 * IPMIコマンド用のメッセージ領域を確保し、メッセージを生成する
 */
static int bmcd_ipmi_alloc_msg(bmcd_cmd_t command, void *data, void *sram)
{
	struct kernel_ipmi_msg *msg = &ipmi_msg_buf;
	int rtn = BMCD_NORMAL_END;

	if (data == NULL) {
		switch (command) {
			case BMCD_CMD_SYSSTS_REPORT:
			case BMCD_CMD_SET_ERRLOG:
			case BMCD_CMD_DUMP_UPLOAD_READY:
				BMCD_LOG(BMCD_ERROR, "Invalid parameter, param is null.\n");
				return BMCD_ABNORMAL_END;
			default:
				break;
		}
	}
	if (sram == NULL) {
		switch (command) {
			case BMCD_CMD_SET_ERRLOG:
			case BMCD_CMD_DUMP_UPLOAD_READY:
			case BMCD_CMD_DUMP_UPLOAD_START:
				BMCD_LOG(BMCD_ERROR, "Invalid parameter, sram is null.\n");
				return BMCD_ABNORMAL_END;
			default:
				break;
		}
	}

	switch (command) {
		case BMCD_CMD_SYSSTS_REPORT:
			msg->netfn		= 0x38;
			msg->cmd		= 0x01;
			bmcd_ipmi_data_set(msg, data, sizeof(struct bmcd_ipmi_param_report));
			break;

		case BMCD_CMD_SHUTDOWN_REASON:
			msg->netfn		= 0x38;
			msg->cmd		= 0x02;
			bmcd_ipmi_data_set(msg, NULL, 0);
			break;

		case BMCD_CMD_SET_ERRLOG:
			msg->netfn		= 0x38;
			msg->cmd		= 0x31;
			bmcd_ipmi_data_set(msg, data, sizeof(struct bmcd_ipmi_param_errlog));
			rtn = bmcd_ipmi_sram_set(sram, ((struct bmcd_ipmi_param_errlog *)data)->sram_size);
			break;

		case BMCD_CMD_DUMP_UPLOAD_READY:
			msg->netfn		= 0x38;
			msg->cmd		= 0x40;
			bmcd_ipmi_data_set(msg, data, sizeof(struct bmcd_ipmi_param_dump1));
			rtn = bmcd_ipmi_sram_set(sram, ((struct bmcd_ipmi_param_dump1 *)data)->sram_size);
			break;

		case BMCD_CMD_DUMP_UPLOAD_START:
			msg->netfn		= 0x38;
			msg->cmd		= 0x41;
			bmcd_ipmi_data_set(msg, (void *)&serial_no, 4);
			rtn = bmcd_ipmi_sram_set(sram, sizeof(struct bmcd_ipmi_sram_dump2));
			serial_no++;
			break;

		case BMCD_CMD_DUMP_UPLOAD_END:
			msg->netfn		= 0x38;
			msg->cmd		= 0x43;
			bmcd_ipmi_data_set(msg, NULL, 0);
			break;

		default:
			BMCD_LOG(BMCD_WARN, "Invalid command, command = %d.\n", command);
			rtn = BMCD_ABNORMAL_END;
			break;
	}

	return rtn;
}

static int bmcd_ipmi_request_settime(struct kernel_ipmi_msg *msg)
{
	int err;
	int rtn = -1;
	int count = 0;

	while (rtn < 0) {
		err = __ipmi_request_settime(ipmi_user, &ipmi_addr_buf, 0, msg, NULL,0,0,0);
		switch (err) {
			case 0:
				rtn = E_NORMALEND;
				break;
			case -ENODEV:
				rtn = E_HARDERROR;
				break;
			case -ENOMEM:
				if (count >= BMCD_IPMI_ENOMEM_RETRY_MAX) {
					rtn = E_BUFFULL;
				}
				count++;
				break;
			case -EINVAL:
			case -EMSGSIZE:
			default:
				rtn = E_INTERNALERROR;
				break;
		}
	}

	if (rtn != E_NORMALEND) {
		BMCD_LOG(BMCD_WARN, "ipmi_request_settime is failed, err = %d.\n", err);
	}

	return rtn;
}

/**
 * IPMIコマンドのリトライ処理を行う。
 * リトライ回数が、第一引数の最大リトライ回数以上の場合は、第二引数の値を返す。
 * 最大リトライ回数未満の場合は、第三引数の時間(ms)waitしてIPMIコマンドを再発行する。
 */
static int bmcd_ipmi_count_retry(int retry_max, int retry_over_rtn, int delay_time)
{
	int error;

	/* リトライオーバーか判定 */
	if (retry_max <= retry_count) {
		BMCD_LOG(BMCD_WARN, "Retry over, retry_max[%d] retry_count[%d].\n",
			retry_max, retry_count);
		return retry_over_rtn;
	}

	retry_count++;
	
	/* 指定時間待ってIPMIコマンド再発行 */
	mdelay(delay_time);
	error = bmcd_ipmi_request_settime(&ipmi_msg_buf);
	if (error != E_NORMALEND) {
		return error;
	}

	BMCD_LOG(BMCD_INFO, "Retry end, retry_max[%d] retry_count[%d].\n", retry_max, retry_count);
	return BMCD_IPMI_RETRY;
}

/**
 * IPMIコマンドのリトライ処理を行う。
 * リトライ処理が3分未満の場合は、10秒waitしてIPMIコマンドを再発行する。
 * 3分以上の場合は、E_COMTIMEOUTを返す。
 */
static int bmcd_ipmi_time_retry(void)
{
	int error;
	struct timeval tv;

	do_gettimeofday(&tv);

	/* 初回のリトライ処理か判定 */
	if (retry_start_time == 0) {
		retry_start_time = tv.tv_sec;
	}

	/* リトライオーバーか判定 */
	if ((retry_start_time + 180) < tv.tv_sec) {
		BMCD_LOG(BMCD_WARN, "Retry over, retry_start_time[%d] tv_sec[%d].\n",
			retry_start_time, tv.tv_sec);
		return E_COMTIMEOUT;
	}

	/* 10秒待ってIPMIコマンド再発行 */
	mdelay(1000 * 10);
	error = bmcd_ipmi_request_settime(&ipmi_msg_buf);
	if (error != E_NORMALEND) {
		return error;
	}

	BMCD_LOG(BMCD_INFO, "Retry end, retry_start_time[%d] tv_sec[%d].\n", retry_start_time, tv.tv_sec);
	return BMCD_IPMI_RETRY;
}

static bmcd_return_t bmcd_ipmi_err_handler(unsigned char comp_code)
{
	bmcd_return_t rtn = E_NORMALEND;

	/**
 	 * Completion Codeを判定し、異常コードの場合はリトライ処理などを行う
 	 */
	switch (comp_code) {
		case 0x00:	/* Command Completed Normally. */
			break;
		case 0xC0:	/* Node busy. */
		case 0xCF:	/* Cannot execute duplicated request. */
		case 0xD0:	/* Command response could not be provided. SDR Repository in update mode. */
		case 0xD1:	/* Command response could not be provided. Device in firmware update mode. */
		case 0xD2:	/* Command response could not be provided. BMC initialization or 
					initialization agent in progress. */
		case 0xD5:	/* Cannot execute command. Command, or request parameter(s), not supported
					in present state. */
			rtn = bmcd_ipmi_count_retry(5, E_BUSY, 1000);
			break;

		case 0xC3:	/* Timeout while processing command. */
			rtn = bmcd_ipmi_time_retry();
			break;

		case 0xC4:	/* Out of space. */
			rtn = bmcd_ipmi_count_retry(10, E_BUFFULL, 500);
			break;

		case 0xC5:	/* Reservation Canceled or Invalid Reservation ID. */
			rtn = E_SYSERROR;
			break;

		case 0xC6:	/* Request data truncated. */
			rtn = bmcd_ipmi_count_retry(5, E_IOERROR, 1000);
			break;

		case 0xC1:	/* Invalid Command. */
		case 0xC2:	/* Command invalid for given LUN. */
		case 0xC7:	/* Request data length invalid. */
		case 0xC8:	/* Request data field length limit exceeded. */
		case 0xC9:	/* Parameter out of range. */
		case 0xCA:	/* Cannot return number of requested data bytes. */
		case 0xCC:	/* Invalid data field in Request. */
		case 0xCD:	/* Command illegal for specified sensor or record type. */
		case 0xFF:	/* Unspecified error. */
			rtn = E_COMERROR;
			break;

		case 0xCB:	/* Requested Sensor, data, or record not present. */
		case 0xCE:	/* Command response could not be provided. */
			rtn = E_HARDERROR;
			break;

		case 0xD3:	/* Destination unavailable. */
			rtn = E_NOTREADY;
			break;

		case 0xD4:	/* Cannot execute command due to insufficient privilege level or
					other security-based restriction. */
		case 0x81:	/* "Dump upload ready notice" is not running. */
			rtn = E_INTERNALERROR;
			break;

		case 0xD6:	/* Cannot execute command. Parameter is illegal because command
						sub-function has been disabled or is unavailable */
			rtn = bmcd_ipmi_count_retry(5, E_COMERROR, 1000);
			break;

		default:
			BMCD_LOG(BMCD_WARN, "Invalid code, completion code = %x\n", comp_code);

			/* Completion Codeが異常な場合は、即座にpanicするか判定する */
			if (do_panic) {
				panic("panic appointment processing start=%d", comp_code);
			}
			else {
				rtn = E_INTERNALERROR;
			}
			break;
	}

	BMCD_LOG(BMCD_INFO, "IPMI-command error handler, Completion Code[%x] rtn[%d]\n"
		, comp_code, rtn);
	return rtn;
}

static void bmcd_ipmi_cmd_resp(struct ipmi_recv_msg *msg, void *user_data)
{
	int rtn;
	char recv_msg[IPMI_MAX_MSG_LENGTH];

	/* 応答を受領 */
	if (msg->msg.data_len) {
		memset(recv_msg, 0, IPMI_MAX_MSG_LENGTH);
		memcpy(recv_msg, msg->msg_data, msg->msg.data_len);
	}

	bmcd_log_ipmimsg(BMCD_RESP, ipmi_msg_buf.netfn, ipmi_msg_buf.cmd,
		recv_msg, msg->msg.data_len);

	__ipmi_free_recv_msg(msg);

	/* 割り込みにより強制終了された場合は応答を破棄し、"STANDBY"に遷移 */
	if (ipmi_sync_param.status_code == ERESTARTSYS) {
		//pr_info("[---] bmcd_queue_kill_end.\n");
		bmcd_queue_kill_end();
		return;
	}

	/* Completion Codeチェック */
	rtn = bmcd_ipmi_err_handler(recv_msg[0]);
	if (rtn != BMCD_IPMI_RETRY) {
		ipmi_sync_param.end = true;
		ipmi_sync_param.status_code = rtn;
		wake_up_interruptible(&ipmi_sync_param.resp_wait);
		BMCD_LOG(BMCD_INFO, "Request a wake up\n");
	}
}

void bmcd_ipmi_destroy_user(void)
{
	__ipmi_destroy_user(ipmi_user);
	ipmi_user = NULL;
}

/**
 * @brief bmcd_ipmi_init
 * @note IPMIユーザと待ちキューを初期化する
 */
int bmcd_ipmi_init(void)
{
	int status;
	unsigned int count = 0;

	/**
	 * ipmi_create_userをコールしてIPMIユーザを作成する
	 * 戻り値により以下の振る舞いを行う
	 *       0：SBCRAMのマッピング、waitキューの初期化、IPMIアドレッシングテーブルの初期化
	 * -ENOMEM：最大10回リトライし、10回を超えた場合はBMCD_ABNORMAL_ENDを上位に返す
	 * -ENODEV：BMCD_ABNORMAL_ENDを上位に返す
	 * -EINVAL：BMCD_ABNORMAL_ENDを上位に返す
	 */
	while (1) {
		status = __ipmi_create_user(0, &response_callback, NULL, &ipmi_user);
		if (status == 0) {
			break;
		}
		else if (status == -ENOMEM) {
			if (count < BMCD_IPMI_ENOMEM_RETRY_MAX) {
				count++;
				continue;
			}
		}
		BMCD_LOG(BMCD_ERROR, "ipmi_create_user abnormal end[%d].\n", status);
		errno = -ENOMEM;
		return BMCD_ABNORMAL_END;
	}

	/**
	 * IPMIユーザ作成成功時のみ、SBCRAMのマッピング、waitキューの初期化、
	 * IPMIアドレッシングテーブルの初期化を行う
	 */
#ifdef QEMU
	sbcram_iomap_addr = kmalloc(BMCD_IPMI_SBCRAM_SIZE, GFP_ATOMIC);
#else
	sbcram_iomap_addr = ioremap(BMCD_IPMI_SBCRAM_PHYADDR, BMCD_IPMI_SBCRAM_SIZE);
#endif
	if (sbcram_iomap_addr == NULL) {
		BMCD_LOG(BMCD_ERROR, "ioremap is failed.\n");
		BMCD_SYSMSG_ERRR(15, "Base address is missing. address [0x%016lx]\n", BMCD_IPMI_SBCRAM_PHYADDR);
		bmcd_ipmi_destroy_user();
		errno = -EFAULT;
		return BMCD_ABNORMAL_END;
	}

	BMCD_LOG(BMCD_INFO, "SBCRAM virtual address = %p.\n", sbcram_iomap_addr);
	init_waitqueue_head(&ipmi_sync_param.resp_wait);

	ipmi_addr_buf.addr_type = IPMI_SYSTEM_INTERFACE_ADDR_TYPE;
	ipmi_addr_buf.channel = IPMI_BMC_CHANNEL;
	ipmi_addr_buf.data[0] = 0;

	return BMCD_NORMAL_END;
}

/**
 * @brief bmcd_ipmi_exit
 * @note IPMIユーザを破棄する
 */
void bmcd_ipmi_exit(void)
{
	if (ipmi_user) {
		bmcd_ipmi_destroy_user();
	}
	else {
		BMCD_LOG(BMCD_INFO, "User not create.\n");
	}

	if (sbcram_iomap_addr) {
#ifdef QEMU
		kfree(sbcram_iomap_addr);
#else
		iounmap(sbcram_iomap_addr);
#endif
	}
	else {
		BMCD_LOG(BMCD_INFO, "SBCRAM not mapping.\n");
	}
}

/**
 * @brief bmcd_ipmi_send
 * @param IPMIコマンドコード
 * @param IPMIコマンドフィールドデータ
 * @param IPMIコマンドSRAMデータ
 * @retval BMCドライバ共通戻り値
 * 		E_NORMALEND	：正常終了
 * 		E_INTERNALERROR	：内部論理エラー
 * 		E_HARDERROR	：ハードエラー
 * 		E_BUFFULL	：メモリ枯渇エラー
 * 		ERESTARTSYS	：割り込みによる強制終了
 * @note IPMIコマンドをBMCへ発行する
 */
bmcd_return_t bmcd_ipmi_send(bmcd_cmd_t command, void *data, void *sram)
{
	int rtn;
	int oops_in_progress_save = 0;

	/* IPMIコマンドパラメータ生成 */
	rtn = bmcd_ipmi_alloc_msg(command, data, sram);
	if (rtn == BMCD_ABNORMAL_END) {
		BMCD_LOG(BMCD_WARN, "bmcd_ipmi_alloc_msg is failed.\n");
		return E_INTERNALERROR;
	}

	/* リトライカウンタ・時間・応答待ちフラグリセット */
	retry_count = 0;
	retry_start_time = 0;
	ipmi_sync_param.end = false;

	bmcd_log_ipmimsg(BMCD_REQ, ipmi_msg_buf.netfn, ipmi_msg_buf.cmd,
		ipmi_msg_buf.data, ipmi_msg_buf.data_len);

	/* panic時のoops_in_progressを一時的にoffする */
	if (panic_mode) {
		oops_in_progress_save = oops_in_progress;
		oops_in_progress = 0;
	}

	/* IPMIコマンド発行 */
	rtn = bmcd_ipmi_request_settime(&ipmi_msg_buf);
	if (rtn != E_NORMALEND) {
		BMCD_LOG(BMCD_WARN, "bmcd_ipmi_request_settime is failed, err = %d.\n", rtn);
		return rtn;
	}

	/**
	 * BMCから3分20秒無応答の場合はここでタイムアウトする。
	 * BMCの最大リトライ時間は3分10秒のため、+10秒した3分20秒待てばハード故障による
	 * 無応答と判断した。
	 */
//	pr_info("[++++] wait_event_interruptible_timeout (enter).\n");
	if (!panic_mode) {
		rtn = wait_event_interruptible_timeout(ipmi_sync_param.resp_wait,
							ipmi_sync_param.end == true, BMCD_IPMI_NOTRESP_TIME);
	}
	else {
		unsigned long i;
		rtn = 0;
		for (i = 0; i < BMCD_IPMI_NOTRESP_COUNT; i++) {
			if (ipmi_sync_param.end == true) {
				rtn = 1;
				break;
			}
			__ipmi_poll_interface(ipmi_user);
			mdelay(1);
		}
	}
	if (panic_mode) {
		oops_in_progress = oops_in_progress_save;
	}
//	pr_info("[++++] wait_event_interruptible_timeout (wakeup) = %d.\n",rtn);
	if (rtn > 0) {		/* wakeupした */
		BMCD_LOG(BMCD_INFO, "Wake up, command = %d.\n", command);
		if (ipmi_sync_param.event_type == IE_RESTARTSYS) {
			BMCD_LOG(BMCD_WARN, "Interrupt, command = %d.\n", command);
			ipmi_sync_param.status_code = ERESTARTSYS;
		}
	}
	else if (rtn == 0) {	/* ハード故障による無応答 */
		BMCD_LOG(BMCD_WARN, "Not responding, command = %d.\n", command);
		ipmi_sync_param.status_code = E_HARDERROR;
	}
	else {/*0未満*/		/* 割り込み (恐らく、この割り込みは無い) */
		BMCD_LOG(BMCD_WARN, "Interrupt, command = %d.\n", command);
		ipmi_sync_param.status_code = ERESTARTSYS;
	}

	return ipmi_sync_param.status_code;
}
