/**
 * Copyright(c) 2016-2021 FUJITSU LIMITED All rights reserved.
 */
/*************************************************************
 * include
 *************************************************************/
#include <linux/kernel.h>
//#pragma GCC push_options
//#pragma GCC optimize ("O2")
#include <linux/irq.h>
//#pragma GCC pop_options
#include <linux/kmod.h>		/* call_usermodehelper */
#include <linux/irqdomain.h>
#include <linux/interrupt.h>
#include <linux/acpi.h>
#include <linux/workqueue.h>
#include <linux/cpuhotplug.h>
#include "bmcd_common.h"

/*************************************************************
 * define
 *************************************************************/
#define BMCD_URGENT_IRQ_HW_NO	42
#define BMCD_NORMAL_IRQ_HW_NO	43
#define BMCD_IRQ_UNSET	0

#define BMCD_MMSC_REG_ADDR  (0x001C000000)
#define BMCD_MMSC_REG_SIZE  (64 * 1024)

#define BMCD_MMSC_URGENT_INTERRUPT_ENABLE_OFFSET	(0x0040)
#define BMCD_MMSC_NORMAL_INTERRUPT_ENABLE_OFFSET	(0x0080)
#define BMCD_MMSC_INTERRUPT_ENABLE_SHIFT  (31)
#define BMCD_MMSC_INTERRUPT_ENABLE_MASK   ((1U) << BMCD_MMSC_INTERRUPT_ENABLE_SHIFT)

#define BMCD_MMSC_URGENT_INTERRUPT_STATUS_OFFSET	(0x0044)
#define BMCD_MMSC_NORMAL_INTERRUPT_STATUS_OFFSET	(0x0084)
#define BMCD_MMSC_INTERRUPT_STATUS_SHIFT  (31)
#define BMCD_MMSC_INTERRUPT_STATUS_MASK   ((1U) << BMCD_MMSC_INTERRUPT_STATUS_SHIFT)

#define FX1000_CPUS		(60)
#define GENERIC_A64FX_CPUS 	(48)

/*************************************************************
 * prototype
 *************************************************************/
static int bmcd_hwirq2virq(const int hwirq, bool *registerd);
static void bmcd_panel_req_interrupt_enable(void);
static void bmcd_panel_req_interrupt_disable(void);
static void bmcd_panel_req_status_clear(void);
static irqreturn_t bmcd_normal_irq_handler(int irq, void *dev_id);
static irqreturn_t bmcd_urgent_irq_handler(int irq, void *dev_id);
static void bmcd_irq_shutdown(void);

/*************************************************************
 * global
 *************************************************************/
static int save_urgent_virq = BMCD_IRQ_UNSET;
static int save_normal_virq = BMCD_IRQ_UNSET;
static bool urgent_virq_registered = 0;
static bool normal_virq_registered = 0;
static void __iomem *mmsc_reg_base = NULL;
extern int errno;
static struct workqueue_struct *wq = NULL;
static cpumask_t urgent_cpumask;
static int urgent_bind_jc = -1;
static int urgent_bind_ac = -1;
static enum cpuhp_state save_cpuhp_state = CPUHP_INVALID;
static bool set_urgent_virq = false;

/*************************************************************
 * function
 *************************************************************/
/**
 *  * 物理割り込み番号から論理割り込み番号を取得する
 *   */
static int bmcd_hwirq2virq(const int hwirq, bool *registerd)
{
	int virq = -1, ret = -1;

	ret = acpi_gsi_to_irq(hwirq, &virq);
	if (ret < 0) {
		virq = acpi_register_gsi(NULL, hwirq, ACPI_EDGE_SENSITIVE, ACPI_ACTIVE_HIGH);

		if (virq <= 0) {
			BMCD_LOG(BMCD_ERROR, "acpi_register_gsi abnormal end[%d].\n", virq);
		}
		else {
			if (registerd) {
				*registerd = 1;
			}
		}
	}
	return virq;
}

static void bmcd_do_cpuhp_set_affinity(unsigned int cpu, int set_cpu)
{
	if (cpu == urgent_bind_jc) {
		int ret;
		cpumask_t mask;

		if (unlikely(!cpu_online(set_cpu))) {
			BMCD_LOG(BMCD_ERROR,
				"Interrupt bind target cpu is offline. "
				"(set_cpu=%d, cpu_online_mask=%*pbl)\n",
				set_cpu, cpumask_pr_args(cpu_online_mask));
			/* Error through because cpuhotplug itself fails. */
			return;
		}

		cpumask_clear(&mask);
		cpumask_set_cpu(set_cpu, &mask);

		ret = irq_set_affinity_hint(save_urgent_virq, &mask);
		if (ret) {
			BMCD_LOG(BMCD_ERROR,
				"irq_set_affinity_hint abnormal end[%d].\n",
				ret);
			/* Error through because cpuhotplug itself fails. */
			return;
		}
	}
}

static int bmcd_cpuhp_cpu_online(unsigned int cpu)
{
	bmcd_do_cpuhp_set_affinity(cpu, urgent_bind_jc);
	return 0;
}

static int bmcd_cpuhp_cpu_offline(unsigned int cpu)
{
	bmcd_do_cpuhp_set_affinity(cpu, urgent_bind_ac);
	return 0;
}

/**
 * @brief bmcd_irq_entry
 * @retval 正常終了：BMCD_NORMAL_END
 *         異常終了：BMCD_ABNORMAL_END
 * @note 割り込み登録を行う
 */
int bmcd_irq_entry(void)
{
	int ret;
	int num_cores;
	unsigned int cpu;

	/* 物理割り込み番号から論理割り込み番号を取得する */
	save_normal_virq = bmcd_hwirq2virq(BMCD_NORMAL_IRQ_HW_NO, &normal_virq_registered);
	if (save_normal_virq <= 0) {
		BMCD_LOG(BMCD_ERROR, "bmcd_hwirq2virq abnormal end. [normal]\n");
		errno = -EINTR;
		goto out_return_err;
	}
	BMCD_LOG(BMCD_INFO, "HW IRQ[%d] , Virtual IRQ[%d]\n", BMCD_NORMAL_IRQ_HW_NO, save_normal_virq);

	save_urgent_virq = bmcd_hwirq2virq(BMCD_URGENT_IRQ_HW_NO, &urgent_virq_registered);
	if (save_urgent_virq <= 0) {
		BMCD_LOG(BMCD_ERROR, "bmcd_hwirq2virq abnormal end. [urgent]\n");
		errno = -EINTR;
		goto out_normal_acpi_unregister;
	}
	BMCD_LOG(BMCD_INFO, "HW IRQ[%d] , Virtual IRQ[%d]\n", BMCD_URGENT_IRQ_HW_NO, save_urgent_virq);

 	/* 割り込みハンドラの登録を行う */
	ret = request_irq(save_normal_virq, bmcd_normal_irq_handler, 0,
			 "MMSC_Normal_interrupt", NULL);
	if (ret) {
		BMCD_LOG(BMCD_ERROR, "request_irq abnormal end[%d]. [normal]\n", ret);
		errno = -EINTR;
		goto out_urgent_acpi_unregister;
	}

	set_urgent_virq = false;
	if (!irq_has_action(save_urgent_virq)) {
		ret = request_irq(save_urgent_virq, bmcd_urgent_irq_handler, 0,
				"MMSC_Urgent_Interruppt", NULL);
		if (ret) {
			BMCD_LOG(BMCD_ERROR, "request_irq abnormal end[%d]. [urgent]\n", ret);
			errno = -EINTR;
			goto out_free_normal_irq;
		}
		set_urgent_virq = true;
	}
	else {
		BMCD_LOG(BMCD_DEBUG, "request_irq : IRQ already entered. ret[%d]. save_urgent_virq[%d]. \n", ret, save_urgent_virq);
	}

	num_cores = num_possible_cpus();
	urgent_bind_jc = num_cores - 2;

	if (num_cores == FX1000_CPUS) {
		int prev_cpu = 0;

		for_each_possible_cpu(cpu) {
			if (!cpu_online(cpu)) {
				urgent_bind_ac = prev_cpu;
				break;
			}
			else {
				prev_cpu = cpu;
			}
		}
	}
	else if (num_cores > 0) { /* GENERIC_A64FX */
		urgent_bind_ac = 0;
	}

	if (unlikely((urgent_bind_ac == -1))) {
		BMCD_LOG(BMCD_ERROR,
			"Loading into an unsupported environment. "
			"(possible_cpu=%d, cpu_online_mask=%*pbl)\n",
			num_cores, cpumask_pr_args(cpu_online_mask));
		errno = -ENODEV;
		goto out_free_urgent_irq;
	}

	if (unlikely(!cpu_online(urgent_bind_jc))) {
		BMCD_LOG(BMCD_ERROR,
			"Interrupt bind target cpu is offline. "
			"(urgent_bind_jc=%d, cpu_online_mask=%*pbl)\n",
			urgent_bind_jc, cpumask_pr_args(cpu_online_mask));
		errno = -ENODEV;
		goto out_free_urgent_irq;
	}

	cpumask_clear(&urgent_cpumask);
	cpumask_set_cpu(urgent_bind_jc, &urgent_cpumask);

	/* パネルリクエストの割り込み先コアを設定する */
	ret = irq_set_affinity_hint(save_urgent_virq, &urgent_cpumask);
	if (ret) {
		BMCD_LOG(BMCD_ERROR, "irq_set_affinity_hint abnormal end[%d].\n", ret);
		errno = ret;
		goto out_free_urgent_irq;
	}

	ret = cpuhp_setup_state_nocalls(CPUHP_AP_ONLINE_DYN,
					"bmc:online",
					bmcd_cpuhp_cpu_online,
					bmcd_cpuhp_cpu_offline);
	if (ret < 0) {
		BMCD_LOG(BMCD_ERROR,
			"cpuhp_setup_state_nocalls abnormal end[%d].\n",
			ret);
		errno = ret;
		goto out_urgent_irq_exit;
	}
	save_cpuhp_state = ret;

	/* ioremap MMSC_REG */
	mmsc_reg_base = ioremap(BMCD_MMSC_REG_ADDR, BMCD_MMSC_REG_SIZE);
	if (mmsc_reg_base == NULL) {
		BMCD_LOG(BMCD_ERROR, "ioremap for MMSC_REG abnormal end.\n");
		errno = -EFAULT;
		goto out_remove_cpuhp;
	}

	/* パネルリクエストの割り込みの status を clear する */
	bmcd_panel_req_status_clear();

	/* パネルリクエストの割り込みを enable にする */
	bmcd_panel_req_interrupt_enable();

	/* シャットダウン処理用のWQを作成する */
	wq = create_singlethread_workqueue("bmc_log_work");
	if (!wq) {
		BMCD_LOG(BMCD_ERROR, "create_singlethread_workqueue error.\n");
		errno = -EIO;
		goto out_iounmap;
	}

	return BMCD_NORMAL_END;

out_iounmap:
	if (mmsc_reg_base) {
		iounmap(mmsc_reg_base);
		mmsc_reg_base = NULL;
	}

out_remove_cpuhp:
	cpuhp_remove_state_nocalls(save_cpuhp_state);

out_urgent_irq_exit:
	cpumask_copy(&urgent_cpumask, cpu_possible_mask);
	ret = irq_set_affinity_hint(save_urgent_virq, &urgent_cpumask);
	if (ret) {
	        BMCD_LOG(BMCD_ERROR, "irq_set_affinity_hint abnormal end[%d].\n", ret);
	}

out_free_urgent_irq:
	if (set_urgent_virq) {
		free_irq(save_urgent_virq, NULL);
	}

out_free_normal_irq:
	free_irq(save_normal_virq, NULL);

out_urgent_acpi_unregister:
	if (urgent_virq_registered) {
		acpi_unregister_gsi(BMCD_URGENT_IRQ_HW_NO);
		urgent_virq_registered = 0;
	}

out_normal_acpi_unregister:
	if (normal_virq_registered) {
		acpi_unregister_gsi(BMCD_NORMAL_IRQ_HW_NO);
		normal_virq_registered = 0;
	}

out_return_err:
	return BMCD_ABNORMAL_END;
}

/**
 * パネルリクエストの割り込みを enable にする
 */
static void bmcd_panel_req_interrupt_enable(void)
{
	u32 mmsc;
	const void __iomem *normal_enable_reg_addr;
	const void __iomem *urgent_enable_reg_addr;

	/* MMSC_Normal_Interruppt_Enable レジスタを read する */
	normal_enable_reg_addr = mmsc_reg_base + BMCD_MMSC_NORMAL_INTERRUPT_ENABLE_OFFSET;
	mmsc = readl_relaxed(normal_enable_reg_addr);

	/* disable であれば、enable にする */
	if (!(mmsc & BMCD_MMSC_INTERRUPT_ENABLE_MASK)) {
		mmsc |= ((1U) << BMCD_MMSC_INTERRUPT_ENABLE_SHIFT);
		writel_relaxed(mmsc, (void*)normal_enable_reg_addr);
	}

	/* MMSC_Urgent_Interruppt_Enable レジスタを read する */
	urgent_enable_reg_addr = mmsc_reg_base + BMCD_MMSC_URGENT_INTERRUPT_ENABLE_OFFSET;
	mmsc = readl_relaxed(urgent_enable_reg_addr);

	/* disable であれば、enable にする */
	if (!(mmsc & BMCD_MMSC_INTERRUPT_ENABLE_MASK)) {
		mmsc |= ((1U) << BMCD_MMSC_INTERRUPT_ENABLE_SHIFT);
		writel_relaxed(mmsc, (void*)urgent_enable_reg_addr);
	}
}

/**
 * パネルリクエストの割り込みを disable にする
 */
static void bmcd_panel_req_interrupt_disable(void)
{
	u32 mmsc;
	const void __iomem *normal_enable_reg_addr;
	const void __iomem *urgent_enable_reg_addr;

	/* MMSC_Normal_Interruppt_Enable レジスタを read する */
	normal_enable_reg_addr = mmsc_reg_base + BMCD_MMSC_NORMAL_INTERRUPT_ENABLE_OFFSET;
	mmsc = readl_relaxed(normal_enable_reg_addr);

	/* enable であれば、disable にする */
	if (mmsc & BMCD_MMSC_INTERRUPT_ENABLE_MASK) {
		mmsc &= ~BMCD_MMSC_INTERRUPT_ENABLE_MASK;
		writel_relaxed(mmsc, (void*)normal_enable_reg_addr);
	}

	/* MMSC_Urgent_Interruppt_Enable レジスタを read する */
	urgent_enable_reg_addr = mmsc_reg_base + BMCD_MMSC_URGENT_INTERRUPT_ENABLE_OFFSET;
	mmsc = readl_relaxed(urgent_enable_reg_addr);

	/* enable であれば、disable にする */
	if (mmsc & BMCD_MMSC_INTERRUPT_ENABLE_MASK) {
		mmsc &= ~BMCD_MMSC_INTERRUPT_ENABLE_MASK;
		writel_relaxed(mmsc, (void*)urgent_enable_reg_addr);
	}
}

/**
 * パネルリクエストの割り込みの status を clear する
 */
static void bmcd_panel_req_status_clear(void)
{
	u32 mmsc;
	const void __iomem *normal_status_reg_addr;
	const void __iomem *urgent_status_reg_addr;

	/* MMSC_Normal_Interruppt_Status レジスタを read する */
	normal_status_reg_addr = mmsc_reg_base + BMCD_MMSC_NORMAL_INTERRUPT_STATUS_OFFSET;
	mmsc = readl_relaxed(normal_status_reg_addr);

	/* pending 中であれば、clear する */
	if (mmsc & BMCD_MMSC_INTERRUPT_STATUS_MASK) {
		/* 1 を書き込むことで clear されるビット */
		mmsc |= ((1U) << BMCD_MMSC_INTERRUPT_STATUS_SHIFT);
		writel_relaxed(mmsc, (void*)normal_status_reg_addr);
	}

	/* MMSC_Urgent_Interruppt_Status レジスタを read する */
	urgent_status_reg_addr = mmsc_reg_base + BMCD_MMSC_URGENT_INTERRUPT_STATUS_OFFSET;
	mmsc = readl_relaxed(urgent_status_reg_addr);

	/* pending 中であれば、clear する */
	if (mmsc & BMCD_MMSC_INTERRUPT_STATUS_MASK) {
		/* 1 を書き込むことで clear されるビット */
		mmsc |= ((1U) << BMCD_MMSC_INTERRUPT_STATUS_SHIFT);
		writel_relaxed(mmsc, (void*)urgent_status_reg_addr);
	}
}

/**
 * @brief bmcd_irq_free
 * @note 割り込みを解除する
 */
void bmcd_irq_free(void)
{
	int ret;

	if (wq) {
		flush_workqueue(wq);
		destroy_workqueue(wq);
		wq = NULL;
	}

	bmcd_panel_req_interrupt_disable();

	bmcd_panel_req_status_clear();

	if (mmsc_reg_base) {
		iounmap(mmsc_reg_base);
		mmsc_reg_base = NULL;
	}

	cpuhp_remove_state_nocalls(save_cpuhp_state);

	cpumask_copy(&urgent_cpumask, cpu_possible_mask);
	ret = irq_set_affinity_hint(save_urgent_virq, &urgent_cpumask);
	if (ret) {
	        BMCD_LOG(BMCD_ERROR, "irq_set_affinity_hint abnormal end[%d].\n", ret);
	}

	if (set_urgent_virq) {
		free_irq(save_urgent_virq, NULL);
	}

	free_irq(save_normal_virq, NULL);

	if (urgent_virq_registered) {
		acpi_unregister_gsi(BMCD_URGENT_IRQ_HW_NO);
		urgent_virq_registered = 0;
	}

	if (normal_virq_registered) {
		acpi_unregister_gsi(BMCD_NORMAL_IRQ_HW_NO);
		normal_virq_registered = 0;
	}
}

static void bmcd_irq_work_handler(struct work_struct *data)
{
	struct work_struct *work = (struct work_struct *)data;
	kfree(work);
	bmcd_irq_shutdown();
}


/**
 * Normal時の割り込み実行時にコールされ、Get Shutdown Reasonコマンドを発行する。
 */
static irqreturn_t bmcd_normal_irq_handler(int irq, void *dev_id)
{
	struct work_struct *work;

	if (!wq) {
		goto out;
	}

	work = (struct work_struct *)kmalloc(sizeof(struct work_struct), GFP_ATOMIC);
	if (!work) {
		goto out;
	}

	INIT_WORK((struct work_struct *)work, bmcd_irq_work_handler);
	queue_work(wq, (struct work_struct *)work);
out:
	return IRQ_HANDLED;
}

/**
 * @brief bmcd_irq_shutdown
 * @note コマンドの応答結果がE_NORMALENDであれば、Shutdownコマンドを実行する
 */
static void bmcd_irq_shutdown(void)
{
	int ret;
	char shutdown_cmd_path[] = "/sbin/shutdown";
	char *shutdown_cmd_arg[] = {"shutdown", "-h", "now", NULL};
	char *shutdown_envp[] = { "HOME=/", "TERM=linux", NULL, };

		BMCD_SYSMSG_INFO(1004, "Shut down the system by the Shutdown notification.\n");

		/* Shutdownコマンド実行 */
		ret = call_usermodehelper(shutdown_cmd_path,
					shutdown_cmd_arg, shutdown_envp, UMH_WAIT_EXEC);
		if (ret < 0) {
			BMCD_SYSMSG_ERRR(16, "Failed to run the shutdown command[%d]\n", ret);
			BMCD_LOG(BMCD_ERROR, "Shutdown command failed[%d].\n", ret);
			panic("shutdown command demand failed.=%d", ret);
		}

	return;
}

/**
 * パネルリクエストによる割り込み時にコールされ、Panicを発生させる
 */
static irqreturn_t bmcd_urgent_irq_handler(int irq, void *dev_id)
{
	panic("Panel request!!");

	return IRQ_HANDLED;
}
