/**
 * Copyright(c) 2016-2020 FUJITSU LIMITED All rights reserved.
 */

/*************************************************************
 * include
 *************************************************************/
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/namei.h>
#include <linux/spinlock.h>
#include <linux/time.h>
#include <linux/workqueue.h>
#include "bmcd_common.h"

/*************************************************************
 * define
 *************************************************************/
#define BMCD_LOG_DIR		"/var/log/bmc/"
#define BMCD_LOG_FILE_IPMIMSG	"ipmi_message"
#define BMCD_LOG_FILE_COMMON	"common"
#define BMCD_LOG_PATH_IPMIMSG	BMCD_LOG_DIR BMCD_LOG_FILE_IPMIMSG
#define BMCD_LOG_PATH_COMMON	BMCD_LOG_DIR BMCD_LOG_FILE_COMMON

#define BMCD_LOG_TIME_SIZE	32
#define BMCD_LOG_LINE_SIZE	256	/* 1行当たりの最大メッセージサイズ */
#define BMCD_LOG_SIZE_DATA	48	/* IPMIメッセージのデータサイズ */
#define BMCD_LOG_FLAG		(O_APPEND | O_RDWR | O_SYNC | O_CREAT)
#define BMCD_LOG_MODE		0600

#define BMCD_LOG_NETFN_OEM	0x38

/* Log message */
#define BMCD_LOG_IPMIMSG_HEAD1 \
	"     Data/Time      |I/O |     Command name(NetFn:CMD)    |                  Data(0-15)\n"

#define BMCD_LOG_IPMIMSG_HEAD2 \
	"---------------------------------------------------------------------------------------------------------\n"

/*************************************************************
 * variable
 *************************************************************/
int panic_mode = 0;	/* Panic時は1となり、ログ出力を行わない */
static struct file * fp_ipmimsg	= NULL;	/* IPMIメッセージ用ログファイル */
static struct file * fp_common	= NULL;	/* 開発用共通ログファイル */

static bmcd_log_level_t valid_log_level = BMCD_LOG_LEVEL_3;
static spinlock_t bmcd_log_lock;

static struct workqueue_struct *wq = NULL;

/*************************************************************
 * function
 *************************************************************/
/**
 * @brief bmcd_log_level_set
 * @param ログレベル
 * @note ログレベルを設定する
 */
void bmcd_log_level_set(bmcd_log_level_t level)
{
	unsigned long flag;

	if ((level == BMCD_LOG_LEVEL_1) || (level == BMCD_LOG_LEVEL_2) ||
		(level == BMCD_LOG_LEVEL_3) || (level == BMCD_LOG_LEVEL_4)) {
		bmcd_log_level_t local_log_level;

		spin_lock_irqsave(&bmcd_log_lock, flag);
		local_log_level = valid_log_level;
		valid_log_level = level;
		spin_unlock_irqrestore(&bmcd_log_lock, flag);

		BMCD_LOG(BMCD_INFO, "Set the log level. before[%d] after[%d]\n", local_log_level, level);
	}
	else {
		BMCD_LOG(BMCD_WARN, "Unknown log level. level[%d]\n", level);
	}
}

/**
 * @brief bmcd_log_level_get
 * @retval ログレベル
 * @note ログレベルを取得する
 */
bmcd_log_level_t bmcd_log_level_get(void)
{
	int level;
	unsigned long flag;

	spin_lock_irqsave(&bmcd_log_lock, flag);
	level = valid_log_level ;
	spin_unlock_irqrestore(&bmcd_log_lock, flag);

	return level;
}

#ifndef DMESG
static void bmcd_log_work_handler(struct work_struct *data)
{
	log_work_struct_t *log_work = (log_work_struct_t *)data;
	kernel_write(log_work->fp, log_work->msg, strlen(log_work->msg), &(log_work->fp->f_pos));
	kfree(log_work->msg);
	kfree(log_work);
}
#endif

static void bmcd_log_write(struct file *fp, char *s_msg)
{
#ifndef DMESG
	log_work_struct_t *log_work;
	char *msg;

	if (panic_mode) {
		return;
	}

	if (!fp) {
		return;
	}

	if (!wq) {
		return;
	}

	log_work = (log_work_struct_t *)kmalloc(sizeof(log_work_struct_t), GFP_ATOMIC);
	if (!log_work) {
		return;
	}

	msg = (char *)kmalloc(strlen(s_msg)+1, GFP_ATOMIC);
	if (!msg) {
		kfree(log_work);
		return;
	}

	strcpy(msg, s_msg);
	log_work->msg = msg;
	log_work->fp = fp;
	INIT_WORK((struct work_struct *)log_work, bmcd_log_work_handler);
	queue_work(wq, (struct work_struct *)log_work);
#else
	printk(s_msg);
#endif
}

/**
 * @brief bmcd_log_close
 * @note ログファイルをクローズする
 */
void bmcd_log_close(void)
{
	if (fp_ipmimsg) {
		filp_close(fp_ipmimsg, NULL);
		fp_ipmimsg = NULL;
	}

	if (fp_common) {
		filp_close(fp_common, NULL);
		fp_common = NULL;
	}

	if (wq) {
		flush_workqueue(wq);
		destroy_workqueue(wq);
		wq = NULL;
	}
}

/**
 * @brief bmcd_log_open
 * @retval BMCドライバ内部共通戻り値
 * 		BMCD_NORMAL_END			：正常終了
 * 		BMCD_LOG_OPEN_FAILED_IPMIMSG	：IPMIメッセージログオープン失敗
 * 		BMCD_LOG_OPEN_FAILED_COMMON	：共通ログオープン失敗
 * @note ログファイルをオープンする
 */
int bmcd_log_open(void)
{
	struct file *fp_tmp;
	struct path path;
	int not_found;
	int result;

	spin_lock_init(&bmcd_log_lock);

	/* ワーカースレッド起動 */
	wq = create_singlethread_workqueue("bmc_log_work");	
	if (!wq) {
		return BMCD_LOG_CREATE_WORKQUEUE;
	}

	/* IPMI message file */
	fp_tmp = filp_open(BMCD_LOG_PATH_IPMIMSG, BMCD_LOG_FLAG, BMCD_LOG_MODE);
	result = IS_ERR(fp_tmp);
	if (result) {
		bmcd_log_close();
		return BMCD_LOG_OPEN_FAILED_IPMIMSG;
	}
	fp_ipmimsg = fp_tmp;

	not_found = kern_path(BMCD_LOG_PATH_IPMIMSG, LOOKUP_CREATE, &path);
	if (not_found) {
		bmcd_log_write(fp_ipmimsg, BMCD_LOG_IPMIMSG_HEAD1);
		bmcd_log_write(fp_ipmimsg, BMCD_LOG_IPMIMSG_HEAD2);
	}

	/* Generic message file */
	fp_tmp = filp_open(BMCD_LOG_PATH_COMMON, BMCD_LOG_FLAG, BMCD_LOG_MODE);
	result = IS_ERR(fp_tmp);
	if (result) {
		bmcd_log_close();
		return BMCD_LOG_OPEN_FAILED_COMMON;
	}
	fp_common = fp_tmp;

	return BMCD_NORMAL_END;
}

static void bmcd_log_vaout(char *out, char *fmt, ...)
{
	va_list args;

	va_start(args, fmt);
	vsnprintf(out, BMCD_LOG_LINE_SIZE, fmt, args);
	va_end(args);
}

static const char * bmcd_log_command(unsigned char netfn, unsigned char cmd)
{
	if (netfn == BMCD_LOG_NETFN_OEM) {
		switch (cmd) {
			case 0x01:
				return "System Status Report(38:01)     ";
			case 0x02:
				return "Get Shutdown Reason(38:02)      ";
			case 0x31:
				return "Set Node OS Error Log(38:31)    ";
			case 0x40:
				return "Dump Upload Ready Notice(38:40) ";
			case 0x41:
				return "Dump Upload Start(38:41)        ";
			case 0x43:
				return "Dump Upload End(38:43)          ";
			default:
				break;
		}
	}
	return "UNKNOWN COMMAND                 ";
}

static const char * bmcd_log_data(const char *data, unsigned int len, char *buf)
{
	int i;

	if ((data == NULL) || (len == 0)) {
		snprintf(buf, BMCD_LOG_SIZE_DATA, "%02d", 0);
	}
	else {
		snprintf(buf, BMCD_LOG_SIZE_DATA, "%02d", data[0]);
		for (i = 1; i < len; i++) {
			snprintf(buf, BMCD_LOG_SIZE_DATA, "%s:%02d", buf, data[i]);
		}
	}

	return buf;
}

static void bmcd_log_time(char *buf)
{
	time64_t cur_time = ktime_get_real_seconds();
	struct tm cur_tm;

	time64_to_tm(cur_time, 0, &cur_tm);

	snprintf(buf, BMCD_LOG_TIME_SIZE, "%ld/%d/%d %d:%d:%d",
		cur_tm.tm_year + 1900, cur_tm.tm_mon + 1, cur_tm.tm_mday,
		cur_tm.tm_hour, cur_tm.tm_min, cur_tm.tm_sec);
}

/**
 * @brief bmcd_log_ipmimsg
 * @note IPMIメッセージを書き込む
 */
void bmcd_log_ipmimsg(bmcd_log_ipmi_t io, unsigned char netfn, unsigned char cmd,
			const char *data, unsigned int data_len)
{
	char log[BMCD_LOG_LINE_SIZE];
	char *io_str[BMCD_LOG_IO_NUM] = {"REQ ", "RESP"};
	char buf[BMCD_LOG_SIZE_DATA];
	char time[BMCD_LOG_TIME_SIZE];

	bmcd_log_time(time);
	bmcd_log_vaout(log, "%s|%s|%s|%s\n",
		time, io_str[io], bmcd_log_command(netfn, cmd), bmcd_log_data(data, data_len, buf));
	bmcd_log_write(fp_ipmimsg, log);
}

/**
 * @brief bmcd_log_print
 * @param ログレベル
 * @param ソースファイルパス
 * @param 行番号
 * @param 出力する文字列
 * @note 開発用共通ログを書き込む
 */
void bmcd_log_print(bmcd_log_common_t level, const char *file, 
			int line, const char *fmt, ...)
{
	static char *level_name[BMCD_LOG_LEVEL_NUM] = {
		"ERR ", "WARN", "INFO", "DEBG"};
	va_list args;
	char log[BMCD_LOG_LINE_SIZE];
	char msg[BMCD_LOG_LINE_SIZE];
	char time[BMCD_LOG_TIME_SIZE];

	/* ログレベルが有効か確認する */
	if (valid_log_level & (1 << level)) {
		va_start(args, fmt);
		vsnprintf(msg, sizeof(msg), fmt, args);
		va_end(args);

		bmcd_log_time(time);
		bmcd_log_vaout(log, "[%s][%s:%d][%s] %s",
			time, bmcd_common_basename(file), line, level_name[level], msg);
		bmcd_log_write(fp_common, log);
	}
}

