/**
 * Copyright(c) 2016-2021 FUJITSU LIMITED All rights reserved.
 */

/*************************************************************
 * include
 *************************************************************/
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/cdev.h>
#include <linux/device.h>
#include <linux/fs.h>
#include <linux/slab.h>
#include <linux/proc_fs.h>
#include <linux/uaccess.h>
#include <linux/delay.h>
#include "bmcd_common.h"

/*************************************************************
 * define
 *************************************************************/
#define DEV_NAME	"bmc_drv"
#define DEV_VERSION	"1.0"
#define DEV_DESCRIPTION	"BMC driver"
#define DEV_COPYRIGHT	"Copyright(c) 2018 FUJITSU LIMITED. All rights reserved."

#define BMCD_PROC_DIR		DEV_NAME
#define BMCD_PROC_FILE_PANIC	"panic"
#define BMCD_PROC_FILE_LOGLEVEL	"log_level"
#define BMCD_PROC_IO_SIZE	3
#define BMCD_REQ_NOTRESP_COUNT	300 * 1000 /* リクエストコマンドポーリング回数 */

extern struct ipmi_sync_param ipmi_sync_param;
extern int panic_mode;

/*************************************************************
 * struct
 *************************************************************/
struct bmcd_dev_t {
	dev_t 			id;
	struct cdev 		*cdev;
	struct proc_dir_entry 	*proc_dir;
	struct proc_dir_entry 	*proc_panic;
	struct proc_dir_entry 	*proc_loglv;
};

struct bmcd_sync_param {
	wait_queue_head_t wait_queue;
	short ret;
	bool end;
};

/*************************************************************
 * global
 *************************************************************/
static struct bmcd_dev_t bmcd_dev = {0};
int errno;

/*************************************************************
 * function(EXPORT)
 *************************************************************/
static void bmcd_sync_callback(short ret, void *param)
{
	struct bmcd_sync_param *sync_param;

	sync_param = (struct bmcd_sync_param *)param;

	sync_param->ret = ret;
	sync_param->end = true;
	if (panic_mode == 0) {
		wake_up_interruptible(&sync_param->wait_queue);
	}
}

static unsigned short bmcd_request(struct bmcd_queue_request *req,
					void(*async_callback)(short))
{
	int rtn;
	unsigned short ret_code = E_NORMALEND;
	struct bmcd_sync_param sync_param;

	if (async_callback) {
		req->async_func = async_callback;
		req->sync_func = NULL;
	}
	else {
		init_waitqueue_head(&sync_param.wait_queue);
		req->private = &sync_param;
		req->async_func = NULL;
		req->sync_func = bmcd_sync_callback;
	}

	/* イベント関連の初期化を行う */
	ipmi_sync_param.end = false;
	ipmi_sync_param.event_type = IE_NORMAL;
	ipmi_sync_param.status_code = 0;
	sync_param.end = false;

	rtn = bmcd_queue_entry(req);
	if (rtn == BMCD_ABNORMAL_END) {
		ret_code = E_INTERNALERROR;
		BMCD_LOG(BMCD_WARN, "bmcd_queue_entry abnormal end.\n");
	}
	else {
		if (req->sync_func) {
			if (panic_mode == 0) {
				//pr_info("[****] wait_event_interruptible (enter).\n");
				rtn = wait_event_interruptible(sync_param.wait_queue, sync_param.end == true);
				//pr_info("[****] wait_event_interruptible (wakeup). = %d\n", rtn);
				BMCD_LOG(BMCD_INFO, "Wake up[%d]\n", rtn);
				if (rtn >= 0) {
					ret_code = sync_param.ret;
				}
				else {
					//pr_info("[****] ipmi_sync_param.status_code = ERESTARTSYS.\n");
					ipmi_sync_param.event_type = IE_RESTARTSYS;
					ipmi_sync_param.end = true;
					wake_up_interruptible(&ipmi_sync_param.resp_wait);
				}
			}
			else {
				unsigned long i;
				ret_code = E_COMTIMEOUT;
				for (i = 0; i < BMCD_REQ_NOTRESP_COUNT; i++) {
					if (sync_param.end == true) {
						ret_code = sync_param.ret;
						break;
					}
					mdelay(1);
				}
			}
		}
	}

	BMCD_LOG(BMCD_INFO, "ret_code[0x%x].\n", ret_code);
	bmcd_sysmsg_result(ret_code, req->command);

	//pr_info("[****] BMC ret_code = %d\n", ret_code);
	return ret_code;
}

/**
 * @brief bmcd_osdump_req
 * @param bmcd_dump_info構造体
 * @param コールバック関数
 * @retval 外部向けBMCドライバ戻り値
 * @note Dump upload ready notice/start/endコマンドを発行する
 */
short bmcd_osdump_req(struct bmcd_dump_info *param,
			void(*async_callback)(short))
{
	struct bmcd_queue_request req;
	short ret;

	panic_mode = 1; /* Panic時はログ出力を停止する */

	BMCD_SYSMSG_INFO(1008, "Started the Emergency dump.\n");

	// 暫定的に正常で返す
	//return 0;

	if (!param) {
		BMCD_SYSMSG_ERRR(3, "Invalid parameter. (NULL)\n");
		return E_COMERROR;
	}

	if ((param->filename == NULL) || (param->info == NULL) ||
		(param->namesize == 0) || (param->namesize > OSDUMP_NAME_SIZE-2) ||
		(param->datasize == 0) || (param->datasize > OSDUMP_DATA_SIZE)) {
		BMCD_SYSMSG_ERRR(2, "Invalid parameter. File Name Size:[%d] Data Size:[%d] Name:[%p] Data:[%p]\n",
			param->namesize, param->datasize, param->filename, param->info);
		return E_COMERROR;
	}

	/* ファイルサイズに終端を含める必要があるため、ファイルサイズを終端分加算する。 */
	param->namesize += 1;	

	req.command = BMCD_CMD_DUMP_UPLOAD_READY;
	req.param = param;
	req.param_size = sizeof(struct bmcd_dump_info);
	req.sram = NULL;
	req.sram_size = 0;
	ret = bmcd_request(&req, async_callback);
	panic_mode = 0; /* Panicモードを解除する */
	return ret;
}
EXPORT_SYMBOL(bmcd_osdump_req);

/**
 * @brief bmcd_setreport
 * @param ステータスコード
 * @param コールバック関数(非同期のみ)
 * @retval BMCドライバ共通戻り値
 * @note System Status Reportコマンドを発行する
 */
short bmcd_setreport(short param, void(*async_callback)(short))
{
	struct bmcd_queue_request req;
	struct bmcd_ipmi_param_report ipmi_param;
	short ret;

	BMCD_SYSMSG_INFO(1010, "Started the notification of State notification.\n");

	memset(&ipmi_param, 0, sizeof(struct bmcd_ipmi_param_report));

	switch (param) {
		case E_SYSRUNINFO:
			ipmi_param.node_state = 0x0A;	/* System Running */
			BMCD_SYSMSG_INFO(1003, "Sub System status Report. (System Running)\n");
			break;
		case E_SHUTDOWNINFO:
			ipmi_param.node_state = 0x02;	/* Shutdown Start */
			ipmi_param.add_info = 0x00;	/* Normal shutdown */
			BMCD_SYSMSG_INFO(1002, "Sub System status Report. (Shutdown start)\n");
			break;
		case E_ABSHUTDOWNINFO:
			ipmi_param.node_state = 0x02;	/* Shutdown Start */
			ipmi_param.add_info = 0x01;	/* Abnormal shutdown */
			break;
		case E_PANICINFO:
			BMCD_SYSMSG_INFO(1001, "Sub System status Report. (Panic)\n");
			ipmi_param.node_state = 0x01;	/* Panic */
			panic_mode = 1;	/* Panic時はログ出力を停止する */
			// 暫定的に正常で返す。
			//return 0;
			break;
		default:
			BMCD_SYSMSG_ERRR(3, "Invalid parameter. [%d]\n", param);
			return E_COMERROR;
			break;
	}
	BMCD_LOG(BMCD_INFO, "param[%d] node_state[%d] add_info[%d].\n",
		param, ipmi_param.node_state, ipmi_param.add_info);

	req.command = BMCD_CMD_SYSSTS_REPORT;
	req.param = &ipmi_param;
	req.param_size = sizeof(struct bmcd_ipmi_param_report);
	req.sram = NULL;
	req.sram_size = 0;
	ret = bmcd_request(&req, async_callback);
	panic_mode = 0; /* Panicモードを解除する */
	return ret;
}
EXPORT_SYMBOL(bmcd_setreport);

/**
 * @brief bmcd_senderrlog
 * @param bmcd_accident_info構造体ポインタ
 * @param コールバック関数(非同期のみ)
 * @retval BMCドライバ共通戻り値
 * @note Set Node OS Error Logコマンドを送信する
 */
short bmcd_senderrlog(struct bmcd_accident_info *param,
			void (*async_callback)(short))
{
	struct bmcd_queue_request req = {0};
	struct bmcd_ipmi_param_errlog ipmi_param;
	struct bmcd_ipmi_sram_errlog *ipmi_sram;
	short rtn;

	BMCD_SYSMSG_INFO(1007, "Started the notification of trouble information.\n");

	if (!param) {
		BMCD_SYSMSG_ERRR(3, "Invalid parameter. (NULL)\n");
		return E_COMERROR;
	}

	if ((param->info == NULL) || (param->size > ERRLOG_DATA_SIZE) ||
		(param->size == 0)) {
		BMCD_SYSMSG_ERRR(4, "Invalid parameter. Info:[%p] Size:[%d]\n",
			param->info, param->size);
		return E_COMERROR;
	}

	ipmi_sram = kmalloc(sizeof(struct bmcd_ipmi_sram_errlog), GFP_ATOMIC);
	if (ipmi_sram == NULL) {
		BMCD_SYSMSG_ERRR(5, "Detected an internal error of the IPMI command.. [38:31]\n");
		return E_INTERNALERROR;
	}

	ipmi_param.sram_size = param->size;
	ipmi_param.sram_check_sum = bmcd_common_checksum(param->info, ipmi_param.sram_size);
	memcpy(ipmi_sram->log, param->info, param->size);

	req.command = BMCD_CMD_SET_ERRLOG;
	req.param = &ipmi_param;
	req.param_size = sizeof(struct bmcd_ipmi_param_errlog);
	req.sram = (void *)ipmi_sram;
	req.sram_size = sizeof(struct bmcd_ipmi_sram_errlog);
	rtn = bmcd_request(&req, async_callback);

	kfree(ipmi_sram);
	return rtn;
}
EXPORT_SYMBOL(bmcd_senderrlog);

/*************************************************************
 * function
 *************************************************************/
static int bmcd_open(struct inode *inode, struct file *filp)
{
	BMCD_LOG(BMCD_DEBUG, "%s call.\n", __func__);
	return BMCD_NORMAL_END;
}

static int bmcd_close(struct inode *inode, struct file * filp)
{
	BMCD_LOG(BMCD_DEBUG, "%s call.\n", __func__);
	return BMCD_NORMAL_END;
}

static void bmcd_dummy_callback(short ret)
{
	BMCD_LOG(BMCD_INFO, "%s call. ret[%d].\n", __func__, ret);
	return ;
}

static int bmcd_alloc_data(uint32_t size, void __user *from, void **to)
{
	uint64_t count = 0;
	void *buf;

	buf = kmalloc(size, GFP_ATOMIC);
	if (buf == NULL) {
		BMCD_SYSMSG_ERRR(5, "Detected an internal error of the IPMI command.. [38:40]\n");
		return E_INTERNALERROR;
	}

	if (from == NULL || from == (void *)-1) {
		goto invalid_para;
	}

	count = copy_from_user(buf, from, size);
	if (count != 0) {
		goto invalid_para;
	}

	*to = buf;
	return E_NORMALEND;

invalid_para:
		BMCD_SYSMSG_ERRR(20, "Taking the user data failed. size [%lld]\n", count);
		kfree(buf);
		return E_COMERROR;
}

static int bmcd_osdump_pre(struct bmcd_dump_info *info, struct bmcd_osdump_data *dump)
{
	memset(info, 0, sizeof(struct bmcd_dump_info));
	info->datasize = dump->datasize;
	info->namesize = dump->namesize;

	if (dump->name_flag == 0) {
		info->filename = dump->name;
	}
	else if (dump->name_flag == 1) {
	}
	else {
		BMCD_SYSMSG_ERRR(30, "Invalid parameter. name_flag [%d]\n", dump->name_flag);
		goto func_err;
	}

	if (dump->data_flag == 0) {
		info->info = dump->data;
	}
	else if (dump->data_flag == 1) {
	}
	else {
		BMCD_SYSMSG_ERRR(31, "Invalid parameter. data_flag [%d]\n", dump->data_flag);
		goto func_err;
	}

	return E_NORMALEND;

func_err:
	return E_COMERROR;
}

static void bmcd_senderrlog_pre(struct bmcd_accident_info *info,
				struct bmcd_errlog_data *log)
{
	memset(info, 0, sizeof(struct bmcd_accident_info));
	info->size = log->datasize;

	if (!log->data_flag) {
		info->info = log->data;
	}
}

static long bmcd_ioctl(struct file *fp, uint32_t cmd, unsigned long arg)
{
	int ret = E_NORMALEND;
	void (*callback_func)(short) = NULL;
	struct bmcd_dump_info info;
	struct bmcd_accident_info einfo;
	struct bmcd_osdump_data *dump = NULL;
	struct bmcd_errlog_data *elog = NULL;

	BMCD_LOG(BMCD_INFO, "%s call. ioctl command = %u.\n", __func__, cmd);

	switch(cmd) {
		case BMCDIOCREPORT:	/* System Status Report */
			ret = bmcd_setreport(arg, NULL);
			break;

		case BMCDIOCOSDUMP:	/* Dump upload */
			ret = bmcd_alloc_data(sizeof(struct bmcd_osdump_data), (void __user *)arg, (void **)&dump);
			if (ret != E_NORMALEND) {
				BMCD_LOG(BMCD_WARN, "bmcd_alloc_data abnormal end[%d].\n", ret);
				break;
			}
			ret = bmcd_osdump_pre(&info, dump);
			if (ret != E_NORMALEND) {
				kfree(dump);
				break;
			}

			if (dump->wait_flag == 0) {
				callback_func = bmcd_dummy_callback;
			}
			else if (dump->wait_flag == 1) {
			}
			else {
				BMCD_SYSMSG_ERRR(32, "Invalid parameter. wait_flag [%d]\n", dump->wait_flag);
				ret = E_COMERROR;
				kfree(dump);
				break;
			}

			ret = bmcd_osdump_req(&info, callback_func);
			kfree(dump);
			break;

		case BMCDIOCERRLOG:	/* Set Node OS Error Log */
			ret = bmcd_alloc_data(sizeof(struct bmcd_errlog_data), (void __user *)arg, (void **)&elog);
			if (ret != E_NORMALEND) {
				BMCD_LOG(BMCD_WARN, "bmcd_alloc_data abnormal end[%d].\n", ret);
				break;
			}
			bmcd_senderrlog_pre(&einfo, elog);

			if (elog->wait_flag == 0) {
				callback_func = bmcd_dummy_callback;
			}
			else if (elog->wait_flag == 1) {
			}
			else {
				BMCD_SYSMSG_ERRR(32, "Invalid parameter. wait_flag [%d]\n", elog->wait_flag);
				ret = E_COMERROR;
				kfree(elog);
				break;
			}

			ret = bmcd_senderrlog(&einfo, callback_func);
			kfree(elog);
			break;

		default:
			BMCD_SYSMSG_ERRR(17, "Requested an unsupported command. [%u]\n", cmd);
			ret = E_COMERROR;
			break;
	}

	BMCD_LOG(BMCD_INFO, "ioctl return = %d.\n", ret);

	return ret;
}

static struct file_operations bmcf_fops_cdev = {
	.owner		= THIS_MODULE,
	.open		= bmcd_open,
	.unlocked_ioctl	= bmcd_ioctl,
	.release	= bmcd_close,
};

static int bmcd_init_chardev(void)
{
	int error;
	struct cdev *cdev_p;
	bmcd_dev.cdev = NULL;

	error = alloc_chrdev_region(&bmcd_dev.id, 0, 1, DEV_NAME);
	if (error) {
		BMCD_SYSMSG_ERRR(11, "Unable to get device number.\n");
		errno = -ENOMEM;
		goto out_return_error;
	}

	cdev_p = cdev_alloc();
	if (cdev_p == NULL) {
		BMCD_SYSMSG_ERRR(12, "Unable to allocate device file.\n");
		errno = -ENOMEM;
		goto out_unregister;
	}

	cdev_init(cdev_p, &bmcf_fops_cdev);
	cdev_p->owner = THIS_MODULE;

	error = cdev_add(cdev_p, bmcd_dev.id, 1);
	if (error) {
		BMCD_SYSMSG_ERRR(13, "Unable to register device file.\n");
		errno = -EIO;
		goto out_cdev_del;
	}

	bmcd_dev.cdev = cdev_p;

	BMCD_LOG(BMCD_INFO, "Character device create, Major[%d] Minor[%d].\n",
		MAJOR(bmcd_dev.id), MINOR(bmcd_dev.id));

	return BMCD_NORMAL_END;

out_cdev_del:
	cdev_del(bmcd_dev.cdev);

out_unregister:
	unregister_chrdev_region(bmcd_dev.id, 1);

out_return_error:
	return BMCD_ABNORMAL_END;
} 

static void bmcd_exit_chardev(void)
{
	if (bmcd_dev.cdev) { 
		cdev_del(bmcd_dev.cdev);
		unregister_chrdev_region(bmcd_dev.id, 1);
	}
}

static uint64_t bmcd_proc_size(size_t size)
{
	if (size > (BMCD_PROC_IO_SIZE - 1)) {
		size = BMCD_PROC_IO_SIZE - 1;
	}
	return size;
}

static uint64_t bmcd_proc_copy_to_user(int offset, char __user *to, char *from)
{
	int len = BMCD_PROC_IO_SIZE - 1;

	if (offset != 0) {
		return 0;
	}

	if (copy_to_user(to, from, len)) {
		BMCD_LOG(BMCD_WARN, "copy_to_user is failed.\n");
		return 0;
	}
	return len;
}

static ssize_t bmcd_proc_panic_write(struct file *file, const char __user *buf,
					size_t size, loff_t *offset)
{
	char strbuf[BMCD_PROC_IO_SIZE];
	int rtn;

	size = bmcd_proc_size(size);
	rtn = copy_from_user(strbuf, buf, size);
	if (rtn != 0) {
		BMCD_LOG(BMCD_WARN, "copy_from_user is failed.\n");
		return -EFAULT;
	}
	strbuf[size] = '\0';

	switch (strbuf[0]) {
		case '0':
			bmcd_ipmi_panic_set(0);
			break;
		case '1':
			bmcd_ipmi_panic_set(1);
			break;
		default:
			BMCD_LOG(BMCD_WARN, "Invalid panic parameter, parameter = %d.\n", strbuf[0]);
			return -EFAULT;
			break;
	}
	BMCD_LOG(BMCD_INFO, "Set this parameter(%d).\n", strbuf[0]);

	return size;
}

static ssize_t bmcd_proc_panic_read(struct file *file, char __user *buf,
					size_t size, loff_t *offset)
{
	int len = BMCD_PROC_IO_SIZE - 1;
	char strbuf[BMCD_PROC_IO_SIZE];

	snprintf(strbuf, len, "%d", bmcd_ipmi_panic_get());
	strbuf[len] = '\0';

	len = bmcd_proc_copy_to_user(*offset, buf, strbuf);
	*offset += len;
	return len;
}

static struct file_operations bmcd_fops_proc_panic = {
	.owner	= THIS_MODULE,
	.write	= bmcd_proc_panic_write,
	.read	= bmcd_proc_panic_read,
};

static ssize_t bmcd_proc_loglv_write(struct file *file, const char __user *buf,
					size_t size, loff_t *offset)
{
	char strbuf[BMCD_PROC_IO_SIZE];
	unsigned long long number;
	bmcd_log_level_t log_level;
	int rtn;

	size = bmcd_proc_size(size);
	rtn = copy_from_user(strbuf, buf, size);
	if (rtn != 0) {
		BMCD_LOG(BMCD_WARN, "copy_from_user is failed.\n");
		return -EFAULT;
	}
	strbuf[size] = '\0';

	/* memo: strbufが10進数でなければ0を返す */
	number = simple_strtoul(strbuf, NULL, 10);
	switch (number) {
		case 1:
			log_level = BMCD_LOG_LEVEL_1;
			break;
		case 2:
			log_level = BMCD_LOG_LEVEL_2;
			break;
		case 3:
			log_level = BMCD_LOG_LEVEL_3;
			break;
		case 4:
			log_level = BMCD_LOG_LEVEL_4;
			break;
		default:
			BMCD_LOG(BMCD_WARN, "Invalid log level, level = %d.\n", number);
			return -EFAULT;
	}

	bmcd_log_level_set(log_level);
	BMCD_LOG(BMCD_INFO, "Set this parameter(%d).\n", log_level);

	return size;
}

static ssize_t bmcd_proc_loglv_read(struct file *file, char __user *buf,
					size_t size, loff_t *offset)
{
	char strbuf[BMCD_PROC_IO_SIZE];
	int len = BMCD_PROC_IO_SIZE - 1;
	bmcd_log_level_t log_level;
	int log_level_user;

	log_level = bmcd_log_level_get();
	for (log_level_user = 0; log_level != 0;) {
		log_level = log_level >> 1;
		log_level_user++;
	}

	snprintf(strbuf, len, "%d", log_level_user);
	strbuf[len] = '\0';

	len = bmcd_proc_copy_to_user(*offset, buf, strbuf);
	*offset += len;
	return len;
}

static struct file_operations bmcd_fops_proc_loglv = {
	.owner	= THIS_MODULE,
	.write	= bmcd_proc_loglv_write,
	.read	= bmcd_proc_loglv_read,
};

static int bmcd_init_proc(void)
{
	struct proc_dir_entry *proc_entry = NULL;

	/* NULLの時のみRemoveするため、明示的にNULLを設定する */
	bmcd_dev.proc_dir   = NULL;
	bmcd_dev.proc_panic = NULL;
	bmcd_dev.proc_loglv = NULL;

	proc_entry = proc_mkdir(BMCD_PROC_DIR, NULL);
	if (proc_entry == NULL) {
		BMCD_SYSMSG_ERRR(18, "Could not Create procfs entry.\n");
		errno = -EIO;
		goto out_return_err;
	}
	bmcd_dev.proc_dir = proc_entry;

	proc_entry = proc_create(BMCD_PROC_FILE_PANIC, (S_IRUSR|S_IWUSR), 
					bmcd_dev.proc_dir, &bmcd_fops_proc_panic);
	if (proc_entry == NULL) {
		BMCD_SYSMSG_ERRR(19, "Could not Create procfs parameter entry. [%s]\n",
			BMCD_PROC_FILE_PANIC);
		errno = -EIO;
		goto out_dir_del;
	}
	bmcd_dev.proc_panic = proc_entry;

	proc_entry = proc_create(BMCD_PROC_FILE_LOGLEVEL, (S_IRUSR|S_IWUSR), 
					bmcd_dev.proc_dir, &bmcd_fops_proc_loglv);
	if (proc_entry == NULL) {
		BMCD_SYSMSG_ERRR(19, "Could not Create procfs parameter entry. [%s]\n",
			BMCD_PROC_FILE_LOGLEVEL);
		errno = -EIO;
		goto out_panic_del;
	}
	bmcd_dev.proc_loglv = proc_entry;

	return BMCD_NORMAL_END;

out_panic_del:
	proc_remove(bmcd_dev.proc_panic);
	bmcd_dev.proc_panic = NULL;

out_dir_del:
	proc_remove(bmcd_dev.proc_dir);
	bmcd_dev.proc_dir = NULL;

out_return_err:
	return BMCD_ABNORMAL_END;
}

static void bmcd_exit_proc(void)
{
	if (bmcd_dev.proc_panic) {
		proc_remove(bmcd_dev.proc_panic);
		bmcd_dev.proc_panic = NULL;
	}

	if (bmcd_dev.proc_loglv) {
		proc_remove(bmcd_dev.proc_loglv);
		bmcd_dev.proc_loglv = NULL;
	}

	if (bmcd_dev.proc_dir) {
		proc_remove(bmcd_dev.proc_dir);
		bmcd_dev.proc_dir = NULL;
	}
}

static int bmcd_init_first(void)
{
	int error;
	
	error = bmcd_ipmi_init();
	if (error) {
		BMCD_SYSMSG_ERRR(26, "Failed to initialize IPMI driver.\n");
		goto out_return_err;
	}

	error = bmcd_irq_entry();
	if (error) {
		BMCD_SYSMSG_ERRR(14, "Unable to register the interrupt handler.\n");
		goto out_ipmi_exit;
	}

	error = bmcd_queue_init();
	if (error) {
		BMCD_SYSMSG_ERRR(25, "Failed to create work queue.\n");
		goto out_irq_free;
	}

	return BMCD_NORMAL_END;

out_irq_free:
	bmcd_irq_free();

out_ipmi_exit:
	bmcd_ipmi_exit();

out_return_err:
	return BMCD_ABNORMAL_END;
}

static int bmcd_init_second(void)
{
	int error;

	error = bmcd_init_chardev();
	if (error) {
		goto out_return_err;
	}

	error = bmcd_init_proc();
	if (error) {
		goto out_cdev_del;
	}

	return BMCD_NORMAL_END;

out_cdev_del:
	bmcd_exit_chardev();

out_return_err:
	return BMCD_ABNORMAL_END;
}

#define READ_ACCESS  (0)
#define WRITE_ACCESS (1)
#define ACCESS_REG_FUNC(name, reg) \
	static void xos_access_##name(uint8_t flag, uint64_t *reg_value){ \
		if(flag == READ_ACCESS){ \
			*reg_value = read_sysreg_s(reg);	\
		} \
		else if(flag == WRITE_ACCESS){ \
			write_sysreg_s(*reg_value, reg); \
		}else{ \
			; \
		}	\
	}

/* Tchipアーキテクチャ判別関数 */
#define XOS_MIDR_EL1		sys_reg(3, 0, 0, 0, 0)
#define XOS_MIDR_EL1_IMPLEMENTER_FJ (0x46)
#define XOS_MIDR_EL1_IMPLEMENTER_MASK (0xFFUL)
#define XOS_MIDR_EL1_IMPLEMENTER_SHIFT (24)
#define XOS_MIDR_EL1_PPNUM_TCHIP (0x1)
#define XOS_MIDR_EL1_PPNUM_MASK (0xFFFUL)
#define XOS_MIDR_EL1_PPNUM_SHIFT (0x4)
ACCESS_REG_FUNC(midr_el1, XOS_MIDR_EL1);

#define XOS_FALSE (0)
#define XOS_TRUE (1)

static int xos_is_tchip(void){

	uint64_t reg = 0;
	int ret=0, impl=0, part=0;

	xos_access_midr_el1(READ_ACCESS, &reg);

	impl = (reg >> XOS_MIDR_EL1_IMPLEMENTER_SHIFT) & XOS_MIDR_EL1_IMPLEMENTER_MASK;
	part = (reg >> XOS_MIDR_EL1_PPNUM_SHIFT) & XOS_MIDR_EL1_PPNUM_MASK;

	if((impl == XOS_MIDR_EL1_IMPLEMENTER_FJ) && (part == XOS_MIDR_EL1_PPNUM_TCHIP)){
		ret = XOS_TRUE;
	}
	else{
		ret = XOS_FALSE;
	}

	return ret;
}

static int __init bmcd_init(void)
{
	int error;

	if (xos_is_tchip() == XOS_FALSE) {
		errno = -ENODEV;
		goto out_return_err;
	}

	BMCD_SYSMSG_INFO(1009, "Started loading BMC driver.\n");
	error = bmcd_log_open();
	switch (error) {
		case BMCD_LOG_OPEN_FAILED_IPMIMSG:
			BMCD_SYSMSG_ERRR(9, "Failed to create a log file for IPMI messages.\n");
			errno = -EIO;
			goto out_return_err;
			break;
		case BMCD_LOG_OPEN_FAILED_COMMON:
			BMCD_SYSMSG_ERRR(8, "Failed to create a log file for developers.\n");
			errno = -EIO;
			goto out_return_err;
			break;
		case BMCD_LOG_CREATE_WORKQUEUE:
			BMCD_SYSMSG_ERRR(25, "Failed to create work queue.\n");
			errno = -EIO;
			goto out_return_err;
			break;
		default:
			break;
	}

	error = bmcd_init_first();
	if (error == BMCD_ABNORMAL_END) {
		goto out_return_err;
	}

	/**
	 * bmcd_init_secondを行うとコマンドを受け付ける状態になるため、
	 * bmcd_init_secondは最後に実行すること。
	 */
	error = bmcd_init_second();
	if (error == BMCD_ABNORMAL_END) {
		goto out_first_exit;
	}

	BMCD_SYSMSG_INFO(1005, DEV_COPYRIGHT);
	BMCD_SYSMSG_INFO(1006, "BMC driver %s: %s.\n", DEV_VERSION, __DATE__);

	return BMCD_NORMAL_END;

out_first_exit:
	bmcd_queue_exit();
	bmcd_irq_free();
	bmcd_ipmi_exit();

out_return_err:
	return errno;
}

static void __exit bmcd_exit(void)
{
	BMCD_SYSMSG_INFO(1014, "Started unloading BMC driver.\n");

	bmcd_exit_proc();
	bmcd_exit_chardev();
	bmcd_queue_exit();
	bmcd_irq_free();
	bmcd_ipmi_exit();
	bmcd_log_close();

	return;
}

module_init(bmcd_init);
module_exit(bmcd_exit);

MODULE_LICENSE("GPL");
MODULE_AUTHOR(DEV_COPYRIGHT);
MODULE_DESCRIPTION(DEV_DESCRIPTION);
