/**
 * Copyright(c) 2016-2020 FUJITSU LIMITED All rights reserved.
 */

/*************************************************************
 * include
 *************************************************************/
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/spinlock.h>
#include <linux/workqueue.h>
#include <linux/delay.h>
#include "bmcd_common.h"

/*************************************************************
 * define
 *************************************************************/
#define BMCD_QUEUE_NAME		"bmcd_queue"
#define BMCD_STATE_RETRY_OVER	182

/*************************************************************
 * global
 *************************************************************/
static bmcd_state_t bmcd_state = BMCD_STATE_UNLOAD;
static spinlock_t bmcd_state_lock;

static struct workqueue_struct *wq = NULL;

extern int errno;

/*************************************************************
 * struct
 *************************************************************/
struct bmcd_queue_work {
	struct work_struct work;
	struct bmcd_queue_request req;
};

/*************************************************************
 * function
 *************************************************************/
static int bmcd_queue_alloc(uint32_t from_size, void *from_ptr,
				uint32_t *to_size, void **to_ptr)
{
	void *buf;

	if ((from_size == 0) || (from_ptr == NULL)) {
		*to_ptr = NULL;
	}
	else {
		buf = kmalloc(from_size, GFP_ATOMIC);
		if (buf == NULL) {
			return BMCD_ABNORMAL_END;
		}
		*to_ptr = buf;
		*to_size = from_size;
		memcpy(*to_ptr, from_ptr, from_size);
	}
	return BMCD_NORMAL_END;
}

static void bmcd_queue_free(struct bmcd_queue_work *work)
{
	if (work->req.sram) {
		kfree(work->req.sram);
	}
	if (work->req.param) {
		kfree(work->req.param);
	}
	kfree(work);
}

static void bmcd_queue_response(bmcd_return_t result, struct bmcd_queue_request *req)
{
	if (req->async_func) {
		req->async_func(result);
	}
	else if (req->sync_func) {
		req->sync_func(result, req->private);
	}
	else {
		BMCD_LOG(BMCD_WARN, "Callback not set. command[%d]\n", req->command);
	}
}

static void bmcd_queue_callback(struct work_struct *work)
{
	int retry_count;
	bool command_destroy = false;
	unsigned long flag;
	bmcd_return_t result = E_BUSY;
	bmcd_cmd_t command;
	struct bmcd_queue_work *bmcd_work = (struct bmcd_queue_work *)work;
	struct bmcd_queue_request *req = &(bmcd_work->req);

	for (retry_count = 0; retry_count < BMCD_STATE_RETRY_OVER; retry_count++) {
		bmcd_state_t local_bmcd_state;

		spin_lock_irqsave(&bmcd_state_lock, flag);

		if (bmcd_state == BMCD_STATE_PROCESSING) {
			spin_unlock_irqrestore(&bmcd_state_lock, flag);
			mdelay(1000);
			BMCD_LOG(BMCD_INFO, "BMC driver is processing state.\n");
			continue;
		}

		command = req->command;

		if (bmcd_state == BMCD_STATE_KILL) {
			spin_unlock_irqrestore(&bmcd_state_lock, flag);
			mdelay(1000);
			BMCD_SYSMSG_INFO(1012, "Could not accept the execution of the command under kill termination. [%d]\n", command);
			continue;
		}

		if (bmcd_state == BMCD_STATE_UNLOAD) {
			spin_unlock_irqrestore(&bmcd_state_lock, flag);
			result = E_NOTREADY;
			BMCD_SYSMSG_ERRR(21, "BMC driver is not ready. [%s]\n",
				bmcd_sysmsg_to_ipmistr(command));
			break;
		}

		bmcd_state = BMCD_STATE_PROCESSING;
		spin_unlock_irqrestore(&bmcd_state_lock, flag);
		BMCD_LOG(BMCD_INFO, "State transitions to processing.\n");

		/* IPMIコマンドを発行する */
		if (command == BMCD_CMD_DUMP_UPLOAD_READY) {
			result = bmcd_dump_ready((struct bmcd_dump_info *)req->param);
		}
		else {
			result = bmcd_ipmi_send(command, req->param, req->sram);
		}

		spin_lock_irqsave(&bmcd_state_lock, flag);
		if (result == ERESTARTSYS) {
			/* 割り込みによる応答のため、状態をKILLに遷移させる */
			bmcd_state = BMCD_STATE_KILL;
		}
		local_bmcd_state = bmcd_state;
		if (bmcd_state == BMCD_STATE_PROCESSING) {
			bmcd_state = BMCD_STATE_STANDBY;
		}
		else {
			/* 状態がPROCESSING以外の時はコマンドを破棄する */
			command_destroy = true;
		}
		spin_unlock_irqrestore(&bmcd_state_lock, flag);

		if (result == ERESTARTSYS) {
			BMCD_LOG(BMCD_INFO, "State transitions to kill.\n");
			BMCD_SYSMSG_INFO(1013, "Command was killed. [%d]\n", command);
		}
		if (local_bmcd_state == BMCD_STATE_PROCESSING) {
			BMCD_LOG(BMCD_INFO, "State transitions to standby.\n");
		}
		else {
			BMCD_LOG(BMCD_WARN, "Current state is not processing.\n");
		}
		break;
	}

	if (command_destroy == false) {
		/* コールバック関数をコールして、コール元に応答する */
		bmcd_queue_response(result, req);
	}

	bmcd_queue_free(bmcd_work);
}

/**
 * @brief bmcd_queue_init
 * @retval 正常終了：BMCD_NORMAL_END
 *         異常終了：BMCD_ABNORMAL_END
 * @note スピンロックの初期化、ステータスの初期化、キューの生成を行う。
 */
int bmcd_queue_init(void)
{
	spin_lock_init(&bmcd_state_lock);
	bmcd_state = BMCD_STATE_STANDBY;

	wq = create_singlethread_workqueue(BMCD_QUEUE_NAME);
	if (wq == NULL) {
		BMCD_LOG(BMCD_ERROR, "create_singlethread_workqueue abnormal end.\n");
		errno = -EIO;
		return BMCD_ABNORMAL_END;
	}

	return BMCD_NORMAL_END;
}

/**
 * @brief bmcd_queue_exit
 * @note キューを破棄する
 */
void bmcd_queue_exit(void)
{
	if (wq) {
		flush_workqueue(wq);
		destroy_workqueue(wq);
	}
}

static int bmcd_queue_set_callback(struct bmcd_queue_request *to,
					struct bmcd_queue_request *from)
{
	if (from->async_func) {
		to->async_func = from->async_func;
		to->sync_func = NULL;
	}
	else if (from->sync_func) {
		to->sync_func = from->sync_func;
		to->async_func = NULL;
	}
	else {
		BMCD_LOG(BMCD_WARN, "Callback not set. command[%d]\n", from->command);
		return BMCD_ABNORMAL_END;
	}
	return BMCD_NORMAL_END;
} 

static int bmcd_queue_do_entry(struct bmcd_queue_request *req, struct bmcd_queue_work *work)
{
	int ret;
	bool direct_call;
	struct bmcd_ipmi_param_report *ipmi_param;

	/* 動的に確保される領域の初期化 */
	work->req.sram = NULL;
	work->req.param = NULL;

	/* IPMIコマンド、プライベートデータを設定 */
	work->req.private = req->private;
	work->req.command = req->command;

	/* IPMIコマンド応答時のコールバック関数を設定 */
	ret = bmcd_queue_set_callback(&(work->req), req);
	if (ret == BMCD_ABNORMAL_END) {
		goto out_queue_free;
	}

	/* SRAMデータを設定 */
	ret = bmcd_queue_alloc(req->sram_size, req->sram, &(work->req.sram_size),
							&(work->req.sram));
	if (ret == BMCD_ABNORMAL_END) {
		BMCD_LOG(BMCD_WARN, "bmcd_queue_alloc abnormal end[sram].\n");
		goto out_queue_free;
	}

	/* IPMIパラメータを設定 */
	ret = bmcd_queue_alloc(req->param_size, req->param, &(work->req.param_size),
							&(work->req.param));
	if (ret == BMCD_ABNORMAL_END) {
		BMCD_LOG(BMCD_WARN, "bmcd_queue_alloc abnormal end[param].\n");
		goto out_queue_free;
	}

	/* 緊急ダンプの場合およびステータス制御がPanicの場合、
　 * ワーカースレッドではなく直接処理を行う */
	direct_call = false;
	if (req->command == BMCD_CMD_DUMP_UPLOAD_READY) {
		direct_call = true;
	}
	else if (req->command == BMCD_CMD_SYSSTS_REPORT) {
		ipmi_param = req->param;
		if (ipmi_param->node_state == 0x01) {	/* Panic */
			direct_call = true;
		}
	}

	if (direct_call) {
		bmcd_queue_callback((struct work_struct*)work);
	}
	else {
		/* ワークキューへエンキュー */
		INIT_WORK((struct work_struct*)work, bmcd_queue_callback);
		ret = queue_work(wq, (struct work_struct*)work);
		if (ret == false) {
			BMCD_LOG(BMCD_WARN, "queue_work abnormal end.\n");
			goto out_queue_free;
		}
	}

	return BMCD_NORMAL_END;

out_queue_free:
	bmcd_queue_free(work);
	return BMCD_ABNORMAL_END;
}

/**
 * @brief bmcd_queue_entry
 * @param コマンドパラメータとコールバック関数
 * @retval 正常終了：BMCD_NORMAL_END
 *         異常終了：BMCD_ABNORMAL_END
 * @note ワークキューにエンキューしてIPMIコマンドを発行する
 */
int bmcd_queue_entry(struct bmcd_queue_request *req)
{
	struct bmcd_queue_work *work;

	/* ワークキューが確保されているかチェック */
	if (wq == NULL) {
		BMCD_LOG(BMCD_WARN, "work queue is null.\n");
		return BMCD_ABNORMAL_END;
	}

	/* ワークキューに設定するパラメータの領域を確保 */
	work = kmalloc(sizeof(struct bmcd_queue_work), GFP_ATOMIC);
	if (work == NULL) {
		BMCD_LOG(BMCD_WARN, "kmalloc abnormal end.\n");
		return BMCD_ABNORMAL_END;
	}

	return bmcd_queue_do_entry(req, work);
}

/**
 * @brief bmcd_queue_kill_end
 * @note 割り込み時にKILLとなっていたステータスをSTANDBYに遷移させる
 */
void bmcd_queue_kill_end(void)
{
	unsigned long flag;
	spin_lock_irqsave(&bmcd_state_lock, flag);
	bmcd_state = BMCD_STATE_STANDBY;
	spin_unlock_irqrestore(&bmcd_state_lock, flag);
	BMCD_LOG(BMCD_INFO, "State transitions to standby.\n");
}
